/* 
 * UAE - The Un*x Amiga Emulator
 *
 * MC68000 emulation generator
 *
 * This is a fairly stupid program that generates a lot of case labels that 
 * can be #included in a switch statement.
 * As an alternative, it can generate functions that handle specific
 * MC68000 instructions, plus a prototype header file and a function pointer
 * array to look up the function for an opcode.
 * Error checking is bad, an illegal table68k file will cause the program to
 * call abort().
 * The generated code is sometimes sub-optimal, an optimizing compiler should 
 * take care of this.
 * 
 * (c) 1995 Bernd Schmidt
 * 
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "config.h"
#include "amiga.h"
#include "readcpu.h"

static int n_braces = 0;

static void start_brace(void)
{
    n_braces++;
    printf("{");
}

static void close_brace(void)
{
    assert (n_braces > 0);
    n_braces--;
    printf("}");
}

static void finish_braces(void)
{
    while (n_braces > 0)
	close_brace();
}

static void pop_braces(int to)
{
    while (n_braces > to)
	close_brace();
}

static void genamode(amodes mode, char *reg, wordsizes size, char *name, bool getv, bool movem)
{
    start_brace ();
    switch(mode) {
     case Dreg:
	if (movem)
	    abort();
	if (getv)
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = regs.d[%s];\n", name, reg);
		break;
	     case sz_word:
		printf("\tWORD %s = regs.d[%s];\n", name, reg);
		break;
	     case sz_long:
		printf("\tLONG %s = regs.d[%s];\n", name, reg);
		break;
	     default: abort();
	    }
	break;
     case Areg:
	if (movem)
	    abort();
	if (getv)
	    switch(size) {	  
	     case sz_word:
		printf("\tWORD %s = regs.a[%s];\n", name, reg);
		break;
	     case sz_long:
		printf("\tLONG %s = regs.a[%s];\n", name, reg);
		break;
	     default: abort();
	    }
	break;
     case Aind:
	printf("\tCPTR %sa = regs.a[%s];\n", name, reg);
	if (getv)
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	break;
     case Aipi:
	printf("\tCPTR %sa = regs.a[%s];\n", name, reg);
	switch(size) {
	 case sz_byte:	    
	    if (getv) printf("\tBYTE %s = get_byte(%sa);\n", name, name);
	    if (!movem) {
		start_brace();
		printf("\tregs.a[%s] += (%s == 7) ? 2 : 1;\n", reg, reg);
	    }
	    break;
	 case sz_word:
	    if (getv) printf("\tWORD %s = get_word(%sa);\n", name, name);
	    if (!movem) {
		start_brace();
		printf("\tregs.a[%s] += 2;\n", reg);
	    }
	    break;
	 case sz_long:
	    if (getv) printf("\tLONG %s = get_long(%sa);\n", name, name);
	    if (!movem) {
		start_brace();
		printf("\tregs.a[%s] += 4;\n", reg);
	    }
	    break;
	 default: abort();
	}
	break;
     case Apdi:
	switch(size) {	  
	 case sz_byte:
	    if (!movem) printf("\tregs.a[%s] -= (%s == 7) ? 2 : 1;\n", reg, reg);
	    start_brace();
	    printf("\tCPTR %sa = regs.a[%s];\n", name, reg);
	    if (getv) printf("\tBYTE %s = get_byte(%sa);\n", name, name);
	    break;
	 case sz_word:
	    if (!movem) printf("\tregs.a[%s] -= 2;\n", reg);
	    start_brace();
	    printf("\tCPTR %sa = regs.a[%s];\n", name, reg);
	    if (getv) printf("\tWORD %s = get_word(%sa);\n", name, name);
	    break;
	 case sz_long:
	    if (!movem) printf("\tregs.a[%s] -= 4;\n", reg);
	    start_brace();
	    printf("\tCPTR %sa = regs.a[%s];\n", name, reg);
	    if (getv) printf("\tLONG %s = get_long(%sa);\n", name, name);
	    break;
	 default: abort();
	}
	break;
     case Ad16:
	printf("\tCPTR %sa = regs.a[%s] + (LONG)(WORD)nextiword();\n", name, reg);
	if (getv) 
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	break;
     case Ad8r:
	printf("\tCPTR %sa = regs.a[%s];\n", name, reg);
#if 0
	printf("\tUWORD %sdp = nextiword();\n", name);
	printf("\t%sa += (LONG)(BYTE)(%sdp & 0xFF);\n", name, name);
	start_brace();
	printf("\tULONG %sdpr = %sdp & 0x8000 ? regs.a[(%sdp & 0x7000) >> 12] : regs.d[(%sdp & 0x7000) >> 12];\n", name, name, name, name);
	printf("\tif (!(%sdp & 0x800)) %sdpr = (LONG)(WORD)%sdpr;\n", name, name, name);
	printf("\t%sa += %sdpr;\n", name, name);
#endif
	printf("\t%sa = get_disp_ea (%sa, nextiword());\n", name, name);
	if (getv) {
	    start_brace();
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	}
	break;
     case PC16:
	printf("\tCPTR %sa = m68k_getpc();\n", name);
	printf("\t%sa += (LONG)(WORD)nextiword();\n", name);
	if (getv) {
	    start_brace();
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	}
	break;
     case PC8r:
	printf("\tCPTR %sa = m68k_getpc();\n", name);
#if 0
	printf("\tUWORD %sdp = nextiword();\n", name);
	printf("\t%sa += (LONG)(BYTE)(%sdp & 0xFF);\n", name, name);
	start_brace();
	printf("\tULONG %sdpr = %sdp & 0x8000 ? regs.a[(%sdp & 0x7000) >> 12] : regs.d[(%sdp & 0x7000) >> 12];\n", name, name, name, name);
	printf("\tif (!(%sdp & 0x800)) %sdpr = (LONG)(WORD)%sdpr;\n", name, name, name);
	printf("\t%sa += %sdpr;\n", name, name);
#endif
	printf("\t%sa = get_disp_ea (%sa, nextiword());\n", name, name);
	if (getv) {
	    start_brace();
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	}
	break;
     case absw:
	printf("\tCPTR %sa = (LONG)(WORD)nextiword();\n", name);
	if (getv) 
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	break;
     case absl:
	printf("\tCPTR %sa = nextilong();\n", name);
	if (getv) 
	    switch(size) {	  
	     case sz_byte:
		printf("\tBYTE %s = get_byte(%sa);\n", name, name);
		break;
	     case sz_word:
		printf("\tWORD %s = get_word(%sa);\n", name, name);
		break;
	     case sz_long:
		printf("\tLONG %s = get_long(%sa);\n", name, name);
		break;
	     default: abort();
	    }
	break;
     case imm:
	if (getv) 
	    switch(size) {
	     case sz_byte:
		printf("\tBYTE %s = nextiword();\n", name);
		break;
	     case sz_word:
		printf("\tWORD %s = nextiword();\n", name);
		break;
	     case sz_long:
		printf("\tLONG %s = nextilong();\n", name);
		break;
	     default: abort();
	    }
	break;
     case imm0:
	if (!getv) abort();
	printf("\tBYTE %s = nextiword();\n", name);
	break;
     case imm1:
	if (!getv) abort();
	printf("\tWORD %s = nextiword();\n", name);
	break;
     case imm2:
	if (!getv) abort();
        printf("\tLONG %s = nextilong();\n", name);
	break;
     case immi:
	if (!getv) abort();
	printf("\tULONG %s = %s;\n", name, reg);
	break;
     default: 
	abort();
    }
}

static void genastore(char *from, amodes mode, char *reg, wordsizes size, char *to)
{
    switch(mode) {
     case Dreg:
	switch(size) {	  
	 case sz_byte:
	    printf("\tregs.d[%s] &= ~0xff; regs.d[%s] |= (%s) & 0xff;\n", reg, reg, from);
	    break;
	 case sz_word:
	    printf("\tregs.d[%s] &= ~0xffff; regs.d[%s] |= (%s) & 0xffff;\n", reg, reg, from);
	    break;
	 case sz_long:
	    printf("\tregs.d[%s] = (%s);\n", reg, from);
	    break;
	 default: abort();
	}
	break;
     case Areg:
	switch(size) {	  
	 case sz_word:
	    printf("\tregs.a[%s] = (LONG)(WORD)(%s);\n", reg, from);
	    break;
	 case sz_long:
	    printf("\tregs.a[%s] = (%s);\n", reg, from);
	    break;
	 default: abort();
	}
	break;
     case Aind:
     case Aipi:
     case Apdi:
     case Ad16:
     case Ad8r:
     case absw:
     case absl:
	switch(size) {
	 case sz_byte:
	    printf("\tput_byte(%sa,%s);\n", to, from);
	    break;
	 case sz_word:
	    printf("\tput_word(%sa,%s);\n", to, from);
	    break;
	 case sz_long:
	    printf("\tput_long(%sa,%s);\n", to, from);
	    break;
	 default: abort();
	}
	break;
     case PC16:
     case PC8r:
	switch(size) {
	 case sz_byte:
	    printf("\tput_byte(%sa,%s);\n", to, from);
	    break;
	 case sz_word:
	    if (CPU_LEVEL < 2)
		abort();
	    printf("\tput_word(%sa,%s);\n", to, from);
	    break;
	 case sz_long:
	    if (CPU_LEVEL < 2)
		abort();
	    printf("\tput_long(%sa,%s);\n", to, from);
	    break;
	 default: abort();
	}
	break;
     case imm:
     case imm0:
     case imm1:
     case imm2:
     case immi:
	abort();
	break;
     default: 
	abort();
    }
}

static void genmovemel(UWORD opcode)
{
    char getcode[100];
    int size = table68k[opcode].size == sz_long ? 4 : 2;
    
    if (table68k[opcode].size == sz_long) {	
    	strcpy(getcode, "get_long(srca)");
    } else {	    
    	strcpy(getcode, "(LONG)(WORD)get_word(srca)");
    }
    
    printf("\tUWORD mask = nextiword(), bitmask = mask;\n");
    genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, true);
    start_brace();
    printf("\tint i, bitcnt = 0;\n");
    printf("\tfor(i=0;i<16;i++) { bitcnt += bitmask & 1; bitmask >>= 1; }\n");
    
    printf("\tfor(i=0;i<8;i++) { if (mask & 1) { regs.d[i] = %s; srca += %d; } mask >>= 1; }\n", getcode, size);
    printf("\tfor(i=0;i<8;i++) { if (mask & 1) { regs.a[i] = %s; srca += %d; } mask >>= 1; }\n", getcode, size);
    
    if (table68k[opcode].smode == Aipi)
    	printf("\tregs.a[srcreg] = srca;\n");
}

static void genmovemle(UWORD opcode)
{
    char putcode[100], shiftcode[] = ">>";
    int size = table68k[opcode].size == sz_long ? 4 : 2;
    int mask = 1;
    if (table68k[opcode].size == sz_long) {
    	strcpy(putcode, "put_long(srca,");
    } else {	    
    	strcpy(putcode, "put_word(srca,");
    }
    
    printf("\tUWORD mask = nextiword(), bitmask = mask;\n");
    genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, true);
    start_brace();
    printf("\tint i, bitcnt = 0;\n");
    printf("\tULONG rd[8], ra[8];\n");
    printf("\tfor(i=0;i<16;i++) { bitcnt += bitmask & 1; bitmask >>= 1; }\n");
    printf("\tfor(i=0;i<8;i++) { rd[i] = regs.d[i]; ra[i] = regs.a[i]; }\n");
    if (table68k[opcode].smode == Apdi) {
	printf("\tsrca -= %d*bitcnt;\n", size);
	printf("\tregs.a[srcreg] = srca;\n");
	strcpy(shiftcode, "<<");
	mask = 0x8000;
    }

    printf("\tfor(i=0;i<8;i++) { if (mask & %d) { %s rd[i]); srca += %d; } mask %s= 1; }\n", 
	   mask, putcode, size, shiftcode);
    printf("\tfor(i=0;i<8;i++) { if (mask & %d) { %s ra[i]); srca += %d; } mask %s= 1; }\n",
	   mask, putcode, size, shiftcode);    
}

typedef enum {
    flag_logical, flag_add, flag_sub, flag_cmp, flag_addx, flag_subx, flag_zn
} flagtypes;

static void genflags(flagtypes type, wordsizes size, char *value, char *src, char *dst)
{
    char vstr[100],sstr[100],dstr[100];
    char usstr[100],udstr[100];
    char unsstr[100],undstr[100];
    ULONG mask;
    switch(size) {
     case sz_byte: mask = 0xFF; break;
     case sz_word: mask = 0xFFFF; break;
     case sz_long: mask = 0xFFFFFFFF; break;
     default: abort();
    }

    switch(size) {	    
     case sz_byte:
	strcpy(vstr, "((BYTE)(");
	strcpy(usstr, "((UBYTE)(");
	break;
     case sz_word:
	strcpy(vstr, "((WORD)(");
	strcpy(usstr, "((UWORD)(");
	break;
     case sz_long:
	strcpy(vstr, "((LONG)(");
	strcpy(usstr, "((ULONG)(");
	break;
     default:
	abort();
    }
    strcpy(unsstr, usstr); 

    strcpy(sstr, vstr);
    strcpy(dstr, vstr);
    strcat(vstr, value); strcat(vstr,"))");
    strcat(dstr, dst); strcat(dstr,"))");
    strcat(sstr, src); strcat(sstr,"))");
    
    strcpy(udstr, usstr);
    strcat(udstr, dst); strcat(udstr,"))");
    strcat(usstr, src); strcat(usstr,"))");
    
    strcpy(undstr, unsstr);
    strcat(unsstr, "-");
    strcat(undstr, "~");
    strcat(undstr, dst); strcat(undstr,"))");
    strcat(unsstr, src); strcat(unsstr,"))");
    
    switch(type) {
     case flag_logical:
	printf("\tregs.v = regs.c = 0;\n");
	printf("\tregs.z = %s == 0;\n", vstr);
	printf("\tregs.n = %s < 0;\n", vstr);
	break;
     case flag_zn:
	printf("\tif (%s == 0) regs.z = 1;\n", vstr);
	printf("\tregs.n = %s < 0;\n", vstr);
	break;
     case flag_add:
	start_brace();
	printf("\tbool flgs = %s < 0;\n", sstr);
	printf("\tbool flgo = %s < 0;\n", dstr);
	printf("\tbool flgn = %s < 0;\n", vstr);
	printf("\tregs.z = %s == 0;\n", vstr);
	printf("\tregs.v = (flgs == flgo) && (flgn != flgo);\n");
	printf("\tregs.c = regs.x = %s < %s;\n", undstr, usstr);
	printf("\tregs.n = flgn != 0;\n");
	break;
     case flag_sub:
	start_brace();
	printf("\tbool flgs = %s < 0;\n", sstr);
	printf("\tbool flgo = %s < 0;\n", dstr);
	printf("\tbool flgn = %s < 0;\n", vstr);
	printf("\tregs.z = %s == 0;\n", vstr);
	printf("\tregs.v = (flgs != flgo) && (flgn != flgo);\n");
	printf("\tregs.c = regs.x = %s > %s;\n", usstr, udstr);
	printf("\tregs.n = flgn != 0;\n");
	break;
     case flag_addx:
	start_brace();
	printf("\tbool flgs = %s < 0;\n", sstr);
	printf("\tbool flgo = %s < 0;\n", dstr);
	printf("\tbool flgn = %s < 0;\n", vstr);
	printf("\tif ((flgs && flgo && !flgn) || (!flgs && !flgo && flgn)) regs.v = 1;\n");
	printf("\tif ((flgs && flgo) || (!flgn && (flgo || flgs))) regs.c = regs.x = 1;\n");
	break;
     case flag_subx:
	start_brace();
	printf("\tbool flgs = %s < 0;\n", sstr);
	printf("\tbool flgo = %s < 0;\n", dstr);
	printf("\tbool flgn = %s < 0;\n", vstr);
	printf("\tif ((!flgs && flgo && !flgn) || (flgs && !flgo && flgn)) regs.v = 1;\n");
	printf("\tif ((flgs && !flgo) || (flgn && (!flgo || flgs))) regs.c = regs.x = 1;\n");
	break;
     case flag_cmp:
	start_brace();
	printf("\tbool flgs = %s < 0;\n", sstr);
	printf("\tbool flgo = %s < 0;\n", dstr);
	printf("\tbool flgn = %s < 0;\n", vstr);
	printf("\tregs.z = %s == 0;\n", vstr);
	printf("\tregs.v = (flgs != flgo) && (flgn != flgo);\n");
	printf("\tregs.c = %s > %s;\n", usstr, udstr);
	printf("\tregs.n = flgn != 0;\n");
	break;
    }
}

static void gen_opcode(unsigned long int opcode) 
{
    start_brace ();
    switch (table68k[opcode].plev) {
     case 0: /* not priviledged */
	break;
     case 1: /* unpriviledged only on 68000 */
	if (CPU_LEVEL == 0)
	    break;
     case 2: /* priviledged */
	printf("if (!regs.s) { regs.pc_p--; Exception(8); } else\n");
	start_brace();
	break;
     case 3: /* priviledged if size == word */
	if (table68k[opcode].size == sz_byte)
	    break;
	printf("if (!regs.s) { regs.pc_p--; Exception(8); } else\n");
	start_brace();
	break;
    }
    switch(table68k[opcode].mnemo) {
     case i_OR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	printf("\tsrc |= dst;\n");
	genflags(flag_logical, table68k[opcode].size, "src", "", "");
	genastore("src", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_AND:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	printf("\tsrc &= dst;\n");
	genflags(flag_logical, table68k[opcode].size, "src", "", "");
	genastore("src", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_EOR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	printf("\tsrc ^= dst;\n");
	genflags(flag_logical, table68k[opcode].size, "src", "", "");
	genastore("src", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_ORSR:
	printf("\tMakeSR();\n");
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	if (table68k[opcode].size == sz_byte) {
	    printf("\tsrc &= 0xFF;\n");
	}
	printf("\tregs.sr |= src;\n");
	printf("\tMakeFromSR();\n");
	break;
     case i_ANDSR: 	
	printf("\tMakeSR();\n");
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	if (table68k[opcode].size == sz_byte) {
	    printf("\tsrc |= 0xFF00;\n");
	}
	printf("\tregs.sr &= src;\n");
	printf("\tMakeFromSR();\n");
	break;
     case i_EORSR:
	printf("\tMakeSR();\n");
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	if (table68k[opcode].size == sz_byte) {
	    printf("\tsrc &= 0xFF;\n");
	}
	printf("\tregs.sr ^= src;\n");
	printf("\tMakeFromSR();\n");
	break;
     case i_SUB: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = dst - src;\n");
	genflags(flag_sub, table68k[opcode].size, "newv", "src", "dst");
	genastore("newv", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_SUBA:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_long, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = dst - src;\n");
	genastore("newv", table68k[opcode].dmode, "dstreg", sz_long, "dst");
	break;
     case i_SUBX:
	{
	    int old_brace_level;
	    genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	    genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	    start_brace ();
	    printf("\tbool xflg = regs.x;\n");
	    printf("\tregs.v = regs.c = regs.x = 0;\n");
	    start_brace ();
	    printf("\tULONG newv = dst - src;\n");
	    genflags(flag_subx, table68k[opcode].size, "newv", "src", "dst");
	    /* FIXME: I'm not entirely sure how to handle this */
	    printf("\tif (xflg)\n"); 
	    old_brace_level = n_braces;
	    start_brace ();
	    genflags(flag_subx, table68k[opcode].size, "(newv-1)", "1", "newv");
	    printf("\tnewv--;\n");
	    pop_braces(old_brace_level);
	    
	    genflags(flag_zn, table68k[opcode].size, "newv", "", "");
	    genastore("newv", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	}
	break;
     case i_SBCD:
	/* Let's hope this works... */
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	start_brace ();
	printf("\tUWORD newv = dst - src - regs.x;\n");
	printf("\tif ((newv & 0xF) > 9) newv-=6;\n");
	printf("\tregs.c = regs.x = (newv & 0x1F0) > 0x90;\n");
	printf("\tif (regs.c) newv -= 0x60;\n");
	printf("\tif (newv != 0) regs.z = 0;\n");
	genastore("newv", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_ADD:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = dst + src;\n");
	genflags(flag_add, table68k[opcode].size, "newv", "src", "dst");
	genastore("newv", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_ADDA: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_long, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = dst + src;\n");
	genastore("newv", table68k[opcode].dmode, "dstreg", sz_long, "dst");
	break;
     case i_ADDX:
	{
	    int old_brace_level;
	    genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	    genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	    start_brace ();
	    printf("\tbool xflg = regs.x;\n");
	    printf("\tregs.v = regs.c = regs.x = 0;\n");
	    start_brace ();
	    printf("\tULONG newv = dst + src;\n");
	    genflags(flag_addx, table68k[opcode].size, "newv", "src", "dst");
	    printf("\tif (xflg)\n");
	    old_brace_level = n_braces;
	    start_brace ();
	    genflags(flag_addx, table68k[opcode].size, "(newv+1)", "newv", "1");
	    printf("\tnewv++;\n");
	    pop_braces(old_brace_level);
	    genflags(flag_zn, table68k[opcode].size, "newv", "", "");
	    genastore("newv", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	}
	break;
     case i_ABCD:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	start_brace ();
	printf("\tUWORD newv = src + dst + regs.x;\n");
	printf("\tif ((newv & 0xF) > 9) newv+=6;\n");
	printf("\tregs.c = regs.x = (newv & 0x1F0) > 0x90;\n");
	printf("\tif (regs.c) newv += 0x60;\n");
	printf("\tif (newv != 0) regs.z = 0;\n");
	genastore("newv", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_NEG:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace ();
	printf("\tULONG dst = -src;\n");
	genflags(flag_sub, table68k[opcode].size, "dst", "src", "0");
	genastore("dst",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_NEGX:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace ();
	printf("\tULONG dst = -src-regs.x;\n");
	genflags(flag_sub, table68k[opcode].size, "dst", "src", "0");
	genastore("dst",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_NBCD: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace ();
	printf("\tUWORD newv = - src - regs.x;\n");
	printf("\tif ((newv & 0xF) > 9) newv-=6;\n");
	printf("\tregs.c = regs.x = (newv & 0x1F0) > 0x90;\n");
	printf("\tif (regs.c) newv -= 0x60;\n");
	printf("\tif (newv != 0) regs.z = 0;\n");
	genastore("newv", table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_CLR: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	genflags(flag_logical, table68k[opcode].size, "0", "", "");
	genastore("0",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_NOT: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace ();
	printf("\tULONG dst = ~src;\n");
	genflags(flag_logical, table68k[opcode].size, "dst", "", "");
	genastore("dst",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_TST:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genflags(flag_logical, table68k[opcode].size, "src", "", "");
	break;
     case i_BTST:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	if (table68k[opcode].size == sz_byte)
	    printf("\tsrc &= 7;\n");
	else
	    printf("\tsrc &= 31;\n");
	printf("\tregs.z = !(dst & (1 << src));\n");
	break;
     case i_BCHG:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	if (table68k[opcode].size == sz_byte)
	    printf("\tsrc &= 7;\n");
	else
	    printf("\tsrc &= 31;\n");
	printf("\tregs.z = !(dst & (1 << src));\n");
	printf("\tdst ^= (1 << src);\n");
	genastore("dst", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_BCLR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	if (table68k[opcode].size == sz_byte)
	    printf("\tsrc &= 7;\n");
	else
	    printf("\tsrc &= 31;\n");
	printf("\tregs.z = !(dst & (1 << src));\n");
	printf("\tdst &= ~(1 << src);\n");
	genastore("dst", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_BSET:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	if (table68k[opcode].size == sz_byte)
	    printf("\tsrc &= 7;\n");
	else
	    printf("\tsrc &= 31;\n");
	printf("\tregs.z = !(dst & (1 << src));\n");
	printf("\tdst |= (1 << src);\n");
	genastore("dst", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_CMPM:
     case i_CMP:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = dst - src;\n");
	genflags(flag_cmp, table68k[opcode].size, "newv", "src", "dst");
	break;
     case i_CMPA: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_long, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = dst - src;\n");
	genflags(flag_cmp, sz_long, "newv", "src", "dst");
	break;
     case i_MOVEP: 
	printf("\tfprintf(stderr, \"MOVEP\\n\");\n");
	break;
     case i_MOVE:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", false, false);
	genflags(flag_logical, table68k[opcode].size, "src", "", "");
	genastore("src", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_MOVEA:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", false, false);
	genastore("src", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_MVSR2: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	printf("\tMakeSR();\n");
	genastore("regs.sr", table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_MV2SR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	if (table68k[opcode].size == sz_byte)
	    printf("\tMakeSR();\n\tregs.sr &= 0xFF00;\n\tregs.sr |= src & 0xFF;\n");
	else {		    
	    printf("\tregs.sr = src;\n");
	}
	printf("\tMakeFromSR();\n");
	break;
     case i_SWAP: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace ();
	printf("\tULONG dst = ((src >> 16)&0xFFFF) | ((src&0xFFFF)<<16);\n");
	genflags(flag_logical, table68k[opcode].size, "dst", "", "");
	genastore("dst",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_EXG:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	genastore("dst",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	genastore("src",table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_EXT:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_word: printf("\tULONG dst = (LONG)(BYTE)src;\n"); break;
	 case sz_long: printf("\tULONG dst = (LONG)(WORD)src;\n"); break;
	 default: abort();
	}
	genflags(flag_logical, table68k[opcode].size, "dst", "", "");
	genastore("dst",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_MVMEL:
	genmovemel(opcode);
	break;
     case i_MVMLE:
	genmovemle(opcode);
	break;
     case i_TRAP:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	printf("\tException(src+32);\n");
	break;
     case i_MVR2USP:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	printf("\tregs.usp = src;\n");
	break;
     case i_MVUSP2R: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	genastore("regs.usp", table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_RESET:
	printf("\tcustomreset();\n");
	break;
     case i_NOP:
	break;
     case i_STOP:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	printf("\tregs.sr = src;\n");
	printf("\tMakeFromSR();\n");
	printf("\tm68k_setstopped(1);\n");
	break;
     case i_RTE:
	genamode(Aipi, "7", sz_word, "sr", true, false);
	genamode(Aipi, "7", sz_long, "pc", true, false);
	if (CPU_LEVEL > 0) {
	    genamode(Aipi, "7", sz_word, "format", true, false);
	}
	printf("\tregs.sr = sr; m68k_setpc(pc);\n");
	if (CPU_LEVEL > 0) {
	    printf("\tif ((format & 0xF000) == 0x8000) regs.a[7] += 50;\n");
	}
	printf("\tMakeFromSR();\n");
	break;
     case i_RTD:
	break;
     case i_LINK:
	genamode(Apdi, "7", sz_long, "old", false, false);
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genastore("src", Apdi, "7", sz_long, "old");
	genastore("regs.a[7]", table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "offs", true, false);
	printf("\tregs.a[7] += offs;\n");
	break;
     case i_UNLK:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	printf("\tregs.a[7] = src;\n");
	genamode(Aipi, "7", sz_long, "old", true, false);
	genastore("old", table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_RTS:
	genamode(Aipi, "7", sz_long, "pc", true, false);
	printf("\tm68k_setpc(pc);\n");
	break;
     case i_TRAPV:
	printf("\tif(regs.v) Exception(7);\n");
	break;
     case i_RTR: 
	printf("\tMakeSR();\n");
	genamode(Aipi, "7", sz_word, "sr", true, false);
	genamode(Aipi, "7", sz_long, "pc", true, false);
	printf("\tregs.sr &= 0xFF00; sr &= 0xFF;\n");
	printf("\tregs.sr |= sr; m68k_setpc(pc);\n");
	printf("\tMakeFromSR();\n");
	break;
     case i_JSR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	genamode(Apdi, "7", sz_long, "sp", false, false);
	genastore("m68k_getpc()", Apdi, "7", sz_long, "sp");
	printf("\tm68k_setpc(srca);\n");
	break;
     case i_JMP: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	printf("\tm68k_setpc(srca);\n");
	break;
     case i_BSR:
	printf("\tchar *oldpcp = (char *)regs.pc_p;\n");
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(Apdi, "7", sz_long, "sp", false, false);
	genastore("m68k_getpc()", Apdi, "7", sz_long, "sp");
	printf("\tregs.pc_p = (UWORD *)(oldpcp + (LONG)src);\n");
	break;
     case i_Bcc:
	printf("\tchar *oldpcp = (char *)regs.pc_p;\n");
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	printf("\tif (cctrue(%d)) regs.pc_p = (UWORD *)(oldpcp + (LONG)src);\n", table68k[opcode].cc);
	break;
     case i_LEA:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", false, false);
	genastore("srca", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	break;
     case i_PEA:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	genamode(Apdi, "7", sz_long, "dst", false, false);
	genastore("srca", Apdi, "7", sz_long, "dst");
	break;
     case i_DBcc:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "offs", true, false);
	printf("\tif (!cctrue(%d)) {\n", table68k[opcode].cc);
	printf("\tif (src--) regs.pc_p = (UWORD *)((char *)regs.pc_p + (LONG)offs - 2);\n");
	genastore("src", table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	printf("\t}\n");
	break;
     case i_Scc: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", false, false);
	start_brace ();
	printf("\tint val = cctrue(%d) ? 0xff : 0;\n", table68k[opcode].cc);
	genastore("val",table68k[opcode].smode, "srcreg", table68k[opcode].size, "src");
	break;
     case i_DIVU:
	genamode(table68k[opcode].smode, "srcreg", sz_word, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_long, "dst", true, false);
	printf("\tif(src != 0){\n");
	printf("\tULONG newv = (ULONG)dst / (UWORD)src;\n");
	printf("\tULONG rem = (ULONG)dst %% (UWORD)src;\n");
	/* The N flag appears to be set each time there is an overflow.
	 * Weird. */
	printf("\tif (newv > 0xffff) { regs.v = regs.n = 1; } else\n\t{\n");
	genflags(flag_logical, sz_word, "newv", "", "");
	printf("\tnewv = (newv & 0xffff) | ((ULONG)rem << 16);\n");
	genastore("newv",table68k[opcode].dmode, "dstreg", sz_long, "dst");
	printf("\t}\n");
	printf("\t}\n");
	break;
     case i_DIVS: 
	genamode(table68k[opcode].smode, "srcreg", sz_word, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_long, "dst", true, false);
	printf("\tif(src != 0){\n");
	printf("\tLONG newv = (LONG)dst / (WORD)src;\n");
	printf("\tUWORD rem = (LONG)dst %% (WORD)src;\n");
	printf("\tif ((newv & 0xffff0000) && (newv & 0xffff0000) != 0xffff0000) { regs.v = regs.n = 1; } else\n\t{\n");
	printf("\tif (((WORD)rem < 0) != ((LONG)dst < 0)) rem = -rem;\n");
	genflags(flag_logical, sz_word, "newv", "", "");
	printf("\tnewv = (newv & 0xffff) | ((ULONG)rem << 16);\n");
	genastore("newv",table68k[opcode].dmode, "dstreg", sz_long, "dst");
	printf("\t}\n");
	printf("\t}\n");
	break;
     case i_MULU: 
	genamode(table68k[opcode].smode, "srcreg", sz_word, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_word, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = (ULONG)(UWORD)dst * (ULONG)(UWORD)src;\n");
	genflags(flag_logical, sz_long, "newv", "", "");
	genastore("newv",table68k[opcode].dmode, "dstreg", sz_long, "dst");
#ifdef WANT_SLOW_MULTIPLY
	printf("\tspecialflags |= SPCFLAG_EXTRA_CYCLES;\n");
#endif
	break;
     case i_MULS:
	genamode(table68k[opcode].smode, "srcreg", sz_word, "src", true, false);
	genamode(table68k[opcode].dmode, "dstreg", sz_word, "dst", true, false);
	start_brace ();
	printf("\tULONG newv = (LONG)(WORD)dst * (LONG)(WORD)src;\n");
	genflags(flag_logical, sz_long, "newv", "", "");
	genastore("newv",table68k[opcode].dmode, "dstreg", sz_long, "dst");
#ifdef WANT_SLOW_MULTIPLY
	printf("\tspecialflags |= SPCFLAG_EXTRA_CYCLES;\n");
#endif
	break;
     case i_ASR: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tULONG sign = cmask & val;\n");
	printf("\tcnt &= 63;\n");
	printf("\tregs.v = 0;\n");
	printf("\tif (!cnt) { regs.c = 0; } else {");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tregs.c=regs.x=val&1; val = ((ULONG)val >> 1) | sign;\n");
	printf("\t}}\n\tregs.n = sign != 0;\n");
	printf("\tregs.z = val == 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_ASL:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tULONG sign = cmask & val;\n");
	printf("\tcnt &= 63;\n");
	printf("\tregs.v = 0;\n");
	printf("\tif (!cnt) { regs.c = 0; } else { ");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tregs.c=regs.x=(val&cmask)!=0; val <<= 1;\n");
	printf("\tif ((val&cmask)!=sign)regs.v=1;\n");
	printf("\t}}\n\tregs.n = (val&cmask) != 0;\n");
	printf("\tregs.z = val == 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_LSR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tcnt &= 63;\n");
	printf("\tif (!cnt) { regs.c = 0; } else {");
	printf("\tint carry = 0;\n");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tcarry=val&1; val >>= 1;\n");
	printf("\t}\n");
	printf("\tregs.c = regs.x = carry!=0;\n}\n");
	printf("\tregs.n = (val & cmask) != 0; regs.z = val == 0; regs.v = 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");    
	break;
     case i_LSL:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = 0;\n");
	printf("\tcnt &= 63;\n");
	printf("\tif (!cnt) { regs.c = 0; } else {");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tcarry=val&cmask; val <<= 1;\n");
	printf("\t}\n");
	printf("\tregs.c = regs.x = carry!=0;\n}\n");
	printf("\tregs.n = (val & cmask) != 0; regs.z = val == 0; regs.v = 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_ROL:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = 0;\n");
	printf("\tcnt &= 63;\n");	
	printf("\tif (!cnt) { regs.c = 0; } else {");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tcarry=val&cmask; val <<= 1;\n");
	printf("\tif(carry)  val |= 1;\n");
	printf("\t}\n");
	printf("\tregs.c = carry!=0;\n}\n");
	printf("\tregs.n = (val & cmask) != 0; regs.z = val == 0; regs.v = 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_ROR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = 0;\n");	
	printf("\tcnt &= 63;\n");
	printf("\tif (!cnt) { regs.c = 0; } else {");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tcarry=val&1; val = (ULONG)val >> 1;\n");
	printf("\tif(carry) val |= cmask;\n");
	printf("\t}\n");
	printf("\tregs.c = carry!=0;\n}\n");
	printf("\tregs.n = (val & cmask) != 0; regs.z = val == 0; regs.v = 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_ROXL: 
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = 0;\n");
	printf("\tcnt &= 63;\n");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tcarry=val&cmask; val <<= 1;\n");
	printf("\tif(regs.x) val |= 1;\n");    
	printf("\tregs.x = carry != 0;\n");
	printf("\t}\n");
	printf("\tregs.c = regs.x;\n");
	printf("\tregs.n = (val & cmask) != 0; regs.z = val == 0; regs.v = 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_ROXR:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "cnt", true, false);
	genamode(table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = 0;\n");
	printf("\tcnt &= 63;\n");
	printf("\tfor(;cnt;--cnt){\n");
	printf("\tcarry=val&1; val >>= 1;\n");
	printf("\tif(regs.x) val |= cmask;\n");
	printf("\tregs.x = carry != 0;\n");
	printf("\t}\n");
	printf("\tregs.c = regs.x;\n");
	printf("\tregs.n = (val & cmask) != 0; regs.z = val == 0; regs.v = 0;\n");
	genastore("val", table68k[opcode].dmode, "dstreg", table68k[opcode].size, "data");
	break;
     case i_ASRW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	printf("\tregs.v = 0;\n");
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tULONG sign = cmask & val;\n");
	printf("\tregs.c=regs.x=val&1; val = (val >> 1) | sign;\n");
	printf("\tregs.n = sign != 0;\n");
	printf("\tregs.z = val == 0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_ASLW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	printf("\tregs.v = 0;\n");
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tULONG sign = cmask & val;\n");
	printf("\tregs.c=regs.x=(val&cmask)!=0; val <<= 1;\n");
	printf("\tif ((val&cmask)!=sign)regs.v=1;\n");
	printf("\tregs.n = (val&cmask) != 0;\n");
	printf("\tregs.z = val == 0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_LSRW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	printf("\tint carry = val&1;\n");
	printf("\tcarry=val&1; val >>= 1;\n");
	genflags(flag_logical, table68k[opcode].size, "val", "", "");
	printf("regs.c = regs.x = carry!=0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_LSLW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = val&cmask;\n");
	printf("\tval <<= 1;\n");
	genflags(flag_logical, table68k[opcode].size, "val", "", "");
	printf("regs.c = regs.x = carry!=0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_ROLW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = val&cmask;\n");
	printf("\tval <<= 1;\n");
	printf("\tif(carry)  val |= 1;\n");
	genflags(flag_logical, table68k[opcode].size, "val", "", "");
	printf("regs.c = carry!=0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_RORW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	printf("\tint carry = val&1;\n");
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tval >>= 1;\n");
	printf("\tif(carry) val |= cmask;\n");
	genflags(flag_logical, table68k[opcode].size, "val", "", "");
	printf("regs.c = carry!=0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_ROXLW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tint carry = val&cmask;\n");
	printf("\tval <<= 1;\n");
	printf("\tif(regs.x) val |= 1;\n");
	printf("\tregs.x = carry != 0;\n");
	genflags(flag_logical, table68k[opcode].size, "val", "", "");
	printf("regs.x = regs.c = carry!=0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_ROXRW:
	genamode(table68k[opcode].smode, "srcreg", table68k[opcode].size, "data", true, false);
	start_brace ();
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tUBYTE val = data;\n"); break;
	 case sz_word: printf("\tUWORD val = data;\n"); break;
	 case sz_long: printf("\tULONG val = data;\n"); break;
	 default: abort();
	}
	printf("\tint carry = val&1;\n");
	switch(table68k[opcode].size) {
	 case sz_byte: printf("\tULONG cmask = 0x80;\n"); break;
	 case sz_word: printf("\tULONG cmask = 0x8000;\n"); break;
	 case sz_long: printf("\tULONG cmask = 0x80000000;\n"); break;
	 default: abort();
	}
	printf("\tval >>= 1;\n");
	printf("\tif(regs.x) val |= cmask;\n");
	printf("\tregs.x = carry != 0;\n");
	genflags(flag_logical, table68k[opcode].size, "val", "", "");
	printf("regs.x = regs.c = carry!=0;\n");
	genastore("val", table68k[opcode].smode, "srcreg", table68k[opcode].size, "data");
	break;
     case i_MOVEC2:
	genamode (table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace();
	printf("\tint regno = (src >> 12) & 7;\n");
	printf("\tULONG *regp = src & 0x8000 ? regs.a + regno : regs.d + regno;\n");
	printf("\tm68k_movec2(src & 0xFFF, regp);\n");
	break;
     case i_MOVE2C:
	genamode (table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	start_brace();
	printf("\tint regno = (src >> 12) & 7;\n");
	printf("\tULONG *regp = src & 0x8000 ? regs.a + regno : regs.d + regno;\n");
	printf("\tm68k_move2c(src & 0xFFF, regp);\n");
	break;
     case i_CAS:
	{
	    int old_brace_level;
	    genamode (table68k[opcode].smode, "srcreg", table68k[opcode].size, "src", true, false);
	    genamode (table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	    start_brace();
	    printf("\tint ru = (src >> 6) & 7;\n");
	    printf("\tint rc = src & 7;\n");
	    printf("\tif ((LONG)regs.d[rc] == (LONG)dst)");
	    old_brace_level = n_braces;
	    start_brace ();
	    genastore("(regs.d[ru])",table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst");
	    pop_braces (old_brace_level);
	    printf("else");
	    start_brace ();
	    printf("regs.d[rc] = dst;\n");
	    pop_braces (old_brace_level);
	}
	break;
     case i_DIVL:
	genamode (table68k[opcode].smode, "srcreg", table68k[opcode].size, "extra", true, false);
	genamode (table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	printf("\tm68k_divl(opcode, dst, extra);\n");
	break;
     case i_MULL:
	genamode (table68k[opcode].smode, "srcreg", table68k[opcode].size, "extra", true, false);
	genamode (table68k[opcode].dmode, "dstreg", table68k[opcode].size, "dst", true, false);
	printf("\tm68k_mull(opcode, dst, extra);\n");
	break;
     default:
	abort();
	break;
    };
    finish_braces ();
}

static void generate_func(long int from, long int to)
{
    int illg = 0;
    long int opcode;
    UWORD smsk; 
    UWORD dmsk;    

    printf("#include <stdlib.h>\n");
    printf("#include \"config.h\"\n");
    printf("#include \"amiga.h\"\n");
    printf("#include \"options.h\"\n");
    printf("#include \"memory.h\"\n");
    printf("#include \"custom.h\"\n");    
    printf("#include \"newcpu.h\"\n");
    printf("#include \"cputbl.h\"\n");
    for(opcode=from; opcode < to; opcode++) {
	if (table68k[opcode].mnemo == i_ILLG) {
	    illg++;
	    continue;
	}
	if (table68k[opcode].handler != -1) {
	    continue;
	}
#ifndef HAVE_ONE_GIG_OF_MEMORY_TO_COMPILE_THIS
	smsk = (table68k[opcode].sbtarg ? 255 : 7) << table68k[opcode].spos;
	dmsk = 7 << table68k[opcode].dpos;

	printf("void op_%lx(UWORD opcode)\n{\n", opcode);
	if (table68k[opcode].suse
	    && table68k[opcode].smode != imm  && table68k[opcode].smode != imm0
	    && table68k[opcode].smode != imm1 && table68k[opcode].smode != imm2
	    && table68k[opcode].smode != absw && table68k[opcode].smode != absl
	    && table68k[opcode].smode != PC8r && table68k[opcode].smode != PC16)
	{
	    if (table68k[opcode].spos == -1) {
		printf("\tULONG srcreg = %d;\n", (int)table68k[opcode].sreg);
	    } else {
		printf("\tULONG srcreg = (LONG)(BYTE)((opcode & %d) >> %d);\n", smsk, (int)table68k[opcode].spos);
	    }
	}
	if (table68k[opcode].duse
	    /* Yes, the dmode can be imm, in case of LINK or DBcc */
	    && table68k[opcode].dmode != imm  && table68k[opcode].dmode != imm0
	    && table68k[opcode].dmode != imm1 && table68k[opcode].dmode != imm2
	    && table68k[opcode].dmode != absw && table68k[opcode].dmode != absl) {
	    if (table68k[opcode].dpos == -1) {		
		printf("\tULONG dstreg = %d;\n", (int)table68k[opcode].dreg);
	    } else {
		printf("\tULONG dstreg = (opcode & %d) >> %d;\n", dmsk, (int)table68k[opcode].dpos);
	    }
	}
#else
	printf("void op_%lx(UWORD opcode)\n{\n", opcode);
	if (table68k[opcode].suse)
	    printf("\tULONG srcreg = (LONG)(BYTE)%d;\n", (int)table68k[opcode].sreg);
	if (table68k[opcode].duse)
	    printf("\tULONG dstreg = (LONG)(BYTE)%d;\n", (int)table68k[opcode].dreg);
#endif
	gen_opcode(opcode);
        printf("}\n");
    }

    fprintf (stderr, "%d illegals generated.\n", illg);
}

static void generate_table(void)
{
    int illg = 0;
    long int opcode;
    
    printf("#include <stdlib.h>\n");
    printf("#include \"config.h\"\n");
    printf("#include \"amiga.h\"\n");
    printf("#include \"options.h\"\n");
    printf("#include \"memory.h\"\n");
    printf("#include \"custom.h\"\n");
    printf("#include \"newcpu.h\"\n");
    printf("#include \"cputbl.h\"\n");
    
    printf("cpuop_func *cpufunctbl[65536] = {\n");
    for(opcode=0; opcode < 65536; opcode++) {
	if (table68k[opcode].mnemo == i_ILLG) {
	    printf("op_illg");
	    illg++;
	    goto loop_end;
	}
	if (table68k[opcode].handler != -1) {
	    printf("op_%x", table68k[opcode].handler);
	    goto loop_end;
	}
	printf("op_%lx", opcode);
	
	loop_end:
	if (opcode < 65535) printf(",");
	if ((opcode & 7) == 7) printf("\n");
    }
    printf("\n};\n");
    fprintf (stderr, "%d illegals generated.\n", illg);
    if (get_no_mismatches())
	fprintf(stderr, "%d mismatches.\n", get_no_mismatches());
}

static void generate_smalltable(void)
{
    int illg = 0, first = 1;
    long int opcode;
    
    printf("#include <stdlib.h>\n");
    printf("#include \"config.h\"\n");
    printf("#include \"amiga.h\"\n");
    printf("#include \"options.h\"\n");
    printf("#include \"memory.h\"\n");
    printf("#include \"custom.h\"\n");
    printf("#include \"newcpu.h\"\n");
    printf("#include \"cputbl.h\"\n");
    
    printf("struct cputbl smallcputbl[] = {\n");
    for(opcode=0; opcode < 65536; opcode++) {
	if (table68k[opcode].handler == -1 && table68k[opcode].mnemo != i_ILLG) {
	    if (!first) 
		printf(",\n");
	    first = 0;
	    printf("{ op_%x, %d }", opcode, opcode);
	}
    }
    printf("\n};\n");
}

static void generate_header(void)
{
    int illg = 0;
    long int opcode;
    
    printf("extern void op_illg(UWORD) REGPARAM;\n");
    
    for(opcode=0; opcode < 65536; opcode++) {
	if (table68k[opcode].mnemo == i_ILLG) {
	    illg++;
	    continue;
	}
	if (table68k[opcode].handler != -1)
	    continue;
	    
	printf("extern cpuop_func op_%lx;\n", opcode);
    }
    
    fprintf (stderr, "%d illegals generated.\n", illg);
    if (get_no_mismatches())
	fprintf(stderr, "%d mismatches.\n", get_no_mismatches());
}

int main(int argc, char **argv)
{
    long int range = -1;
    char mode = 'n';
    int i;
    
    if (argc == 2)
    	mode = *argv[1];

    if (argc == 3) {
	range = atoi(argv[2]);
	mode = *argv[1];
    }
    
    read_table68k ();
    do_merges ();
    
    switch(mode) {
     case 'f':
    	generate_func(range * 0x1000, (range + 1) * 0x1000);
	break;
     case 'h':
    	generate_header();
	break;
     case 't':
	generate_table();
	break;
     case 's':
	generate_smalltable();
	break;
     default:
	abort();
    }
    free(table68k);
    return 0;
}
