 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * MC68000 emulation
  *
  * (c) 1995 Bernd Schmidt
  * 
  */

extern bool broken_in;

typedef void cpuop_func(UWORD) REGPARAM;

struct cputbl {
    cpuop_func *handler;
    UWORD opcode;
};

extern struct cputbl smallcputbl[];

extern cpuop_func *cpufunctbl[65536];
extern void op_illg(UWORD) REGPARAM;

typedef char flagtype; 

extern struct regstruct 
{
    ULONG d[8];
    CPTR  a[8],usp;
    UWORD sr;
    flagtype t,s,x,n,z,v,c,stopped;
    int intmask;
    ULONG pc;
    UWORD *pc_p;
    UWORD *pc_oldp;
    
    ULONG vbr,sfc,dfc;
} regs;

extern void MC68000_oldstep(UWORD opcode);

static __inline__ UWORD nextiword(void)
{
    UWORD r = *regs.pc_p++;
    return r;
}

static __inline__ ULONG nextilong(void)
{
    ULONG r = *regs.pc_p++;
    r = (r << 16) + *regs.pc_p++;
    return r;
}

static __inline__ void m68k_setpc(CPTR newpc)
{
    regs.pc = newpc;
    regs.pc_p = regs.pc_oldp = get_real_address(newpc);
}

static __inline__ CPTR m68k_getpc(void)
{
    return regs.pc + ((char *)regs.pc_p - (char *)regs.pc_oldp);
}

static __inline__ void m68k_setstopped(int stop)
{
    regs.stopped = stop;
    if (stop)
    	specialflags |= SPCFLAG_STOP;
}

static __inline__ bool cctrue(const int cc)
{
    switch(cc){
     case 0: return true;                                                    /* T */
     case 1: return false;                                                   /* F */
     case 2: return !regs.c && !regs.z;                                      /* HI */
     case 3: return regs.c || regs.z;                                        /* LS */
     case 4: return !regs.c;                                                 /* CC */
     case 5: return regs.c;                                                  /* CS */
     case 6: return !regs.z;                                                 /* NE */
     case 7: return regs.z;                                                  /* EQ */
     case 8: return !regs.v;                                                 /* VC */
     case 9: return regs.v;                                                  /* VS */
     case 10:return !regs.n;                                                 /* PL */
     case 11:return regs.n;                                                  /* MI */
#if 0
     case 12:return (regs.n && regs.v) || (!regs.n && !regs.v);              /* GE */
     case 13:return (regs.n && !regs.v) || (!regs.n && regs.v);              /* LT */
     case 14:return !regs.z && ((regs.n && regs.v) || (!regs.n && !regs.v)); /* GT */
     case 15:return regs.z || ((regs.n && !regs.v) || (!regs.n && regs.v));  /* LE */
#endif 
     case 12:return regs.n == regs.v;             /* GE */
     case 13:return regs.n != regs.v;             /* LT */
     case 14:return !regs.z && (regs.n == regs.v);/* GT */
     case 15:return regs.z || (regs.n != regs.v); /* LE */
    }
    abort();
    return 0;
}

#if CPU_LEVEL > 1
extern ULONG get_disp_ea (ULONG, UWORD);

#else
static __inline__ ULONG get_disp_ea (ULONG base, UWORD dp)
{
    int reg = (dp >> 12) & 7;
    LONG regd;
    
    if (dp & 0x8000)
	regd = regs.a[reg];
    else
	regd = regs.d[reg];
    if (!(dp & 0x800))
	regd = (LONG)(WORD)regd;
    return base + (BYTE)(dp) + regd;
}
#endif

extern void MakeSR(void);
extern void MakeFromSR(void);
extern void Exception(int);
extern void m68k_move2c(int, ULONG *);
extern void m68k_movec2(int, ULONG *);
extern void m68k_divl (UWORD, ULONG, UWORD);
extern void m68k_mull (UWORD, ULONG, UWORD);
extern void init_m68k (void);
extern void MC68000_step(void);
extern void MC68000_run(void);
extern void MC68000_skip(CPTR);
extern void MC68000_dumpstate(CPTR *);
extern void MC68000_disasm(CPTR,CPTR *,int);
extern void MC68000_reset(void);
